#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import scipy as sp
import matplotlib.pyplot as plt

from ..utils.fisher import fisher_information
from ..utils.plotting import get_labels
from ..core.rqa_res import RQARes

class RecurrenceMatrix:
    '''Recurrence matrix object. Used for Recurrence Quantification Analysis (RQA).
    '''
    def __init__(self,matrix,time,epsilon,m,tau,series = None,value_name=None,value_unit=None,time_name=None,time_unit=None,label=None):
        self.matrix = matrix
        self.time = time
        self.epsilon = epsilon
        self.m = m
        self.tau  =tau
        self.series = series
        self.value_name = value_name
        self.value_unit = value_unit
        self.time_name = time_name
        self.time_unit = time_unit
        self.label = label

    def laplacian_eigenmaps(self,w_size, w_incre):
        '''Function to run regime change detection workflow
        
        Parameters
        ----------
            
        w_size : int
            Window size for the fisher information 
        
        w_incre : int 
            Window increment for the fisher information 

        Returns
        -------

        FI_series : pyleoclim.Series object
        
        '''
        W = self.matrix + 1

        D = np.zeros(W.shape)

        for i in range(len(W)):
            D[i,i] = np.sum(W[:,i])
            
        L = D - W

        _, eigvec = sp.linalg.eigh(L,D)
        
        eig_data = []

        for idx, i in enumerate(self.time):
            eig_data.append([i,eigvec[idx,1],eigvec[idx,2],eigvec[idx,3],eigvec[idx,4]])
            
        time,value = fisher_information(eig_data,w_size,w_incre)
        
        FI_series = RQARes(time,
                            value,
                            time_name=self.time_name,
                            time_unit=self.time_unit,
                            value_name=self.value_name,
                            value_unit=self.value_unit,
                            label=self.label,
                            m=self.m,
                            tau=self.tau,
                            eps=self.epsilon
                            )
        
        return FI_series

    def plot(self,figsize=(8,8),xlabel=None,ylabel=None,title=None,imshow_kwargs=None):
        '''Plotting function for recurrence matrices
        
        Parameters
        ----------
        
        figsize : tuple
            Size of figure

        xlabel : str
            Label on x axis (if time name and units are present they will be used)

        ylabel : str
            Label on y axis (if time name and units are present they will be used)

        title : str
            Title of the plot

        imshow_kwargs : dict
            Dictionary of key word arguments for the imshow method from matplotlib.axes.Axes.imshow

        See also
        --------

        matplotlib.axes.Axes.imshow
        '''

        fig, ax = plt.subplots(figsize = figsize)

        imshow_kwargs={} if imshow_kwargs is None else imshow_kwargs.copy()

        if 'cmap' not in imshow_kwargs:
            imshow_kwargs['cmap'] = 'Greys'

        if 'origin' not in imshow_kwargs:
            imshow_kwargs['origin'] = 'lower'

        if 'extent' not in imshow_kwargs:
            imshow_kwargs['extent'] = [self.time[0],self.time[-1],self.time[0],self.time[-1]]

        if xlabel is None:
            xlabel,_ = get_labels(self)
        
        if ylabel is None:
            ylabel,_ = get_labels(self)
        
        if title is None and self.label is not None:
            title = f'{self.label}'

        ax.set_xlabel(xlabel)

        ax.set_ylabel(ylabel)

        ax.set_title(title)

        ax.imshow(self.matrix,**imshow_kwargs)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import warnings
import itertools

import pyleoclim as pyleo
import numpy as np

from tqdm import tqdm
from pyrqa.time_series import TimeSeries
from pyrqa.settings import Settings
from pyrqa.analysis_type import Classic
from pyrqa.neighbourhood import FixedRadius
from pyrqa.metric import EuclideanMetric
from pyrqa.computation import RQAComputation

from ..core.rqa_res import RQARes
from ..core.time_embedded_series import TimeEmbeddedSeries
from ..core.recurrence_matrix import RecurrenceMatrix
from ..utils.parameters import tau_search

class Series(pyleo.Series):
    '''Ammonyte series object, launching point for most ammonyte analysis.

    Child of pyleoclim.Series, so shares all methods with pyleoclim.Series plus those
    defined here.
    '''

    def embed(self,m,tau=None,):
        '''Function to time delay a time series'''

        value_name = self.value_name
        value_unit = self.value_unit
        time_name = self.time_name
        time_unit = self.time_unit
        label = self.label

        if tau is None:
            tau = tau_search(self)

        values = self.value
        time_axis = self.time[:(-m*tau)]
        
        manifold = np.ndarray(shape = (len(values)-(m*tau),m))

        for idx, _ in enumerate(values):
            if idx < (len(values)-(m*tau)):
                manifold[idx] = values[idx:idx+(m*tau):tau]

        embedded_data = manifold
        embedded_time = time_axis

        return TimeEmbeddedSeries(
            self,m,tau,embedded_data,embedded_time,value_name,value_unit,time_name,
            time_unit,label)

    def determinism(self,window_size,overlap,m,tau,eps):
        '''Calculate determinism of a series

        Note that series must be evenly spaced for this method.
        See interp, bin, and gkernel methods in parent class pyleoclim.Series for details.
        
        Parameters
        ----------
        
        window_size : int
            Size of window to use when calculating recurrence plots for determinism statistic.
            Note this is in units of the time axis.
        
        overlap : int
            Amount of overlap to allow between windows.
            Note this is in units of the time axis.

        m : int
            Embedding dimension to use when performing time delay embedding,
            
        tau : int
            Time delay to use when performing time delay embedding
            
        eps : float
            Size of radius to use to calculate recurrence matrix

        Returns
        -------

        det_series : ammonyte.Series
            Ammonyte.Series object containing time series of the determinism statistic
        '''
       
        series = self
        windows = np.arange(int(min(series.time)),int(max(series.time)),int(overlap/2))

        cutoff_index = -int(window_size/(overlap/2))

        res = []
        window_time = []

        for window in tqdm(windows[:cutoff_index]):
            
            series_slice = series.slice((window,window+window_size))

            window_values = series_slice.value
            time = series_slice.time[int((len(series_slice.time)-1)/2)]

            ts = TimeSeries(window_values,
                            embedding_dimension = m,
                            time_delay=tau)

            settings = Settings(ts,
                                analysis_type=Classic,
                                neighbourhood=FixedRadius(eps),
                                similarity_measure=EuclideanMetric)

            computation = RQAComputation.create(settings,
                                                verbose=False)
            
            result = computation.run()

            window_time.append(time)

            res.append(result.determinism)

        det_series = RQARes(
            time = window_time,
            value = res,
            time_name=series.time_name,
            time_unit=series.time_unit,
            value_name='DET',
            label=series.label,
            m = m,
            tau = tau,
            eps = eps)

        return det_series

    def laminarity(self,window_size,overlap,m,tau,eps):
        '''Calculate laminarity of a series

        Note that series must be evenly spaced for this method.
        See interp, bin, and gkernel methods in parent class pyleoclim.Series for details.
        
        Parameters
        ----------
        
        window_size : int
            Size of window to use when calculating recurrence plots for determinism statistic.
            Note this is in units of the time axis.
        
        overlap : int
            Amount of overlap to allow between windows
            Note this is in units of the time axis.

        m : int
            Embedding dimension to use when performing time delay embedding,
            
        tau : int
            Time delay to use when performing time delay embedding
            
        eps : float
            Size of radius to use to calculate recurrence matrix

        Returns
        -------

        lam_series : ammonyte.Series
            Ammonyte.Series object containing time series of the laminarity statistic
        '''

        series = self
        windows = np.arange(int(min(series.time)),int(max(series.time)),int(overlap/2))

        cutoff_index = -int(window_size/(overlap/2))

        res = []
        window_time = []

        for window in tqdm(windows[:cutoff_index]):
            
            series_slice = series.slice((window,window+window_size))

            window_values = series_slice.value
            time = series_slice.time[int((len(series_slice.time)-1)/2)]

            ts = TimeSeries(window_values,
                            embedding_dimension = m,
                            time_delay=tau)

            settings = Settings(ts,
                                analysis_type=Classic,
                                neighbourhood=FixedRadius(eps),
                                similarity_measure=EuclideanMetric)

            computation = RQAComputation.create(settings,
                                                verbose=False)
            
            result = computation.run()

            window_time.append(time)

            res.append(result.laminarity)

        lam_series = RQARes(
            time=window_time,
            value=res,
            time_name=series.time_name,
            time_unit=series.time_unit,
            value_name='LAM',
            label=series.label,
            m = m,
            tau = tau,
            eps = eps)
        
        return lam_series




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pyleoclim as pyleo
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from .sampling import confidence_interval

__all__ = [
    'bootstrap_fill_plot',
    'bootstrap_scatter_plot'
]

def get_labels(obj):
    '''Function to create plotting labels from object metadata.
    
    Assumes specific object properties specific to ammonyte objects
    
    Parameters
    ----------
    
    obj : ammonyte.TimeEmbeddedSeries, ammonyte.RecurrenceMatrix, ammonyte.RecurrenceNetwork'''

    if obj.time_name is not None:
        if obj.time_unit is not None:
            xlabel = f'{obj.time_name} {obj.time_unit}'
        else:
            xlabel = f'{obj.time_name}'
    else:
        xlabel = None
    
    if obj.value_name is not None:
        if obj.value_unit is not None:
            ylabel = f'{obj.value_name} {obj.value_unit}'
        else:
            ylabel = f'{obj.value_name}'
    else:
        ylabel=None

    return xlabel, ylabel

def bootstrap_fill_plot(series,ax=None,line_color=None,fill_color=None,transition_interval=None,
                        plot_kwargs = None, ci_kwargs = None,background_series=None,background_kwargs = None):
    '''Function for plotting rqa results with confidence bounds
    
    series : pyleoclim.Series
        Pyleoclim series object containin timeseries of the plot in question

    ax : matplotlib.axes object
        Axes to plot on, if None new plot will be generated

    line_color : str, tuple
        String or rgb tuple to use for line color
    
    fill_color : str, tuple
        String or rgb tuple to use for fill color in between lines

    transition_interval : list,tuple
        Upper and lower bound for the transition interval

    plot_kwargs : dict
        Key word arguments for the main plot, see `pyleoclim.Series.plot <https://pyleoclim-util.readthedocs.io/en/latest/core/api.html#series-pyleoclim-series>_ for details

    ci_kwargs : dict
        Key word arguments for calculating the confidence interval. Only to be used if `transition_interval` is not passed. See ammonyte.utils.sampling.confidence_interval for details

    background_series : pyleoclim.Series
        Optional to pass a different series that will be plotted behind the main series plot

    background_kwargs : dict
        Key word arguments for the background plot. If none are passed, the color of the main plot will be re-used 
        and alpha will be set to .2

    Returns
    -------

    fig : matplotlib.figure
        The figure object from matplotlib.
        See `matplotlib.pyplot.figure <https://matplotlib.org/stable/api/figure_api.html>`_ for details.

    ax : matplotlib.axis 
        The axis object from matplotlib. 
        See `matplotlib.axes <https://matplotlib.org/stable/api/axes_api.html>`_ for details.


    See also
    --------

    ammonyte.utils.sampling.confidence_interval

    '''

    if not isinstance(series,pyleo.core.Series):
        raise ValueError('This function requires a pyleoclim.Series object, please reformat and pass again.')

    ci_kwargs = {} if ci_kwargs is None else ci_kwargs.copy()

    series = series.interp(step=1)

    if not transition_interval:
        transition_interval = confidence_interval(series,**ci_kwargs)
    
    if line_color is None:
        line_color=sns.color_palette('colorblind')[0]
    if fill_color is None:
        fill_color=sns.color_palette('colorblind')[0]
    
    #Need to find points in time where the series intersects with the lower and upper confidence boundaries
    value = series.value
    time = series.time
    
    idx_ufill = []
    idx_lfill = []

    upper = max(transition_interval)
    lower = min(transition_interval)

    #Need to find all points that lie above or below our confidence interval
    for idx,v in enumerate(value):
        if v >= upper:
            idx_ufill.append(idx)
        elif v <= lower:
            idx_lfill.append(idx)
    
    idx_sort = np.sort(np.concatenate((idx_ufill,idx_lfill)))
    
    if ax is None:
        fig,ax = plt.subplots(figsize=(12,8))

    if background_series:

        if not isinstance(background_series,pyleo.core.Series):
            raise ValueError('background_series does not appear to be a pyleoclim.Series object')
        
        #Standardize background_series values
        background_series.value -= np.mean(background_series.value)
        background_series.value /= max(background_series.value)

        #Adjust scaling to match that of main series
        background_series.value *= max(series.value - np.mean(series.value))
        background_series.value += np.mean(series.value)

        background_kwargs = {} if background_kwargs is None else background_kwargs.copy()

        if 'alpha' not in background_kwargs:
            background_kwargs['alpha'] = .2
        if 'color' not in background_kwargs:
            background_kwargs['color'] = line_color

        background_series.plot(ax=ax, **background_kwargs)

    series.plot(ax=ax,color=line_color)
        
    ufill_values = np.zeros(len(time)) + upper
    lfill_values = np.zeros(len(time)) + lower
    
    for i in range(len(time)):
        if i in idx_ufill:
            ufill_values[i] = value[i]
        elif i in idx_lfill:
            lfill_values[i] = value[i]
                                   
    ax.fill_between(time,lower,upper,color=fill_color,alpha=.1)
    ax.fill_between(time,upper,ufill_values,color=fill_color)
    ax.fill_between(time,lower,lfill_values,color=fill_color)
        
    if 'fig' in locals():
        return fig, ax
    else:
        return ax

def bootstrap_scatter_plot(series,ax=None,marker_color=None,transition_interval=None, scatter_kwargs= None,
                           ci_kwargs = None,background_series=None,background_kwargs = None):
    '''Function for plotting rqa results with confidence bounds
    
    series : pyleoclim.Series
        Pyleoclim series object containin timeseries of the plot in question

    ax : matplotlib.axes object
        Axes to plot on, if None new plot will be generated

    marker_color : str, tuple
        String or rgb tuple to use for line color

    transition_interval : list,tuple
        Upper and lower bound for the transition interval

    scatter_kwargs : dict
        Key word arguments for the scatter plot

    background_series : pyleoclim.Series
        Optional to pass a different series that will be plotted behind the main series plot

    background_kwargs : dict
        Key word arguments for the background plot. If none are passed, the color of the main plot will be re-used 
        and alpha will be set to .2

    Returns
    -------

    fig : matplotlib.figure
        The figure object from matplotlib.
        See `matplotlib.pyplot.figure <https://matplotlib.org/stable/api/figure_api.html>`_ for details.

    ax : matplotlib.axis 
        The axis object from matplotlib. 
        See `matplotlib.axes <https://matplotlib.org/stable/api/axes_api.html>`_ for details.


    See also
    --------

    ammonyte.utils.sampling.confidence_interval

    '''

    if not isinstance(series,pyleo.core.Series):
        raise ValueError('This function requires a pyleoclim.Series object, please reformat and pass again.')

    ci_kwargs = {} if ci_kwargs is None else ci_kwargs.copy()
    scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs.copy()

    series = series.interp(step=1)

    if not transition_interval:
        transition_interval = confidence_interval(series,**ci_kwargs)
    
    if marker_color is None:
        line_color=sns.color_palette('colorblind')[0]
    
    value = series.value
    time = series.time
    
    idx_sig = []

    upper = max(transition_interval)
    lower = min(transition_interval)

    #Need to find all points that lie above or below our confidence interval
    for idx,v in enumerate(value):
        if v >= upper or v <= lower:
            idx_sig.append(idx)
    
    if ax is None:
        fig,ax = plt.subplots(figsize=(12,8))

    if background_series:

        if not isinstance(background_series,pyleo.core.Series):
            raise ValueError('background_series does not appear to be a pyleoclim.Series object')
        
        #Standardize background_series values
        background_series.value -= np.mean(background_series.value)
        background_series.value /= max(background_series.value)

        #Adjust scaling to match that of main series
        background_series.value *= max(series.value - np.mean(series.value))
        background_series.value += np.mean(series.value)

        background_kwargs = {} if background_kwargs is None else background_kwargs.copy()

        if 'alpha' not in background_kwargs:
            background_kwargs['alpha'] = .2
        if 'color' not in background_kwargs:
            background_kwargs['color'] = line_color

        background_series.plot(ax=ax, **background_kwargs)

    ax.scatter(time[idx_sig],value[idx_sig],c = marker_color, **scatter_kwargs)
        
    if 'fig' in locals():
        return fig, ax
    else:
        return ax


''' Tests for ammonyte.core.recurrence_network
Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
    (certain tests will only work when run from the tests directory, so make sure to run from there!)
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pyleoclim as pyleo
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt

from ..utils.sampling import confidence_interval
from ..utils.fisher import smooth_series

class RQARes(pyleo.Series):
    '''Class for storing the result of various RQA techniques'''

    def __init__(self, time, value, time_name=None, time_unit=None, value_name=None, value_unit=None, label=None, m=None,tau=None,eps=None):
        super().__init__(time,value,time_name,time_unit,value_name,value_unit,label)
        self.m = m
        self.tau = tau
        self.eps = eps

    def smooth(self,block_size):
        '''Function to perform block smoothing on your RQA result
        
        Parameters
        ----------
        
        block_size : int
            Number of points to include in each block
            
        Returns
        -------
        
        smoothed_series : ammonyte.RQARes
            Smoothed version of your original RQARes object
            
        See also
        --------
        
        ammonyte.utils.fisher.smooth_series'''

        if block_size is None:
            block_size = int(len(self.time)/15)
            smoothed_series = smooth_series(self,block_size)
        else:
            smoothed_series = smooth_series(self,block_size)

        return smoothed_series

    def confidence_fill_plot(self,ax=None,line_color=None,fill_color=None,fill_alpha=None,transition_interval=None,xlabel=None,ylabel=None,marker=None,
                     markersize=None,linestyle=None,linewidth=None,alpha=None,label=None,title=None,zorder=None,plot_kwargs=None,ci_kwargs=None,
                     background_series=None,background_kwargs=None,legend=True,lgd_kwargs=None):

        '''Function for plotting rqa results with confidence bounds

        Parameters
        ----------

        ax : matplotlib.axes object
            Axes to plot on, if None new plot will be generated

        line_color : str, tuple
            String or rgb tuple to use for line color
        
        fill_color : str, tuple
            String or rgb tuple to use for fill color in between lines

        fill_alpha : float
            Transparency of the fill, default is .1

        transition_interval : list,tuple
            Upper and lower bound for the transition interval

        marker : str
            e.g., 'o' for dots
            See `matplotlib.markers <https://matplotlib.org/stable/api/markers_api.html>`_ for details

        markersize : float
            the size of the marker

        linestyle : str
            e.g., '--' for dashed line
            See `matplotlib.linestyles <https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html>`_ for details

        linewidth : float
            the width of the line

        alpha : float
            Transparency of the line

        label : str
            the label for the line

        xlabel : str
            the label for the x-axis

        ylabel : str
            the label for the y-axis

        title : str
            the title for the figure

        zorder : int
            The default drawing order for all lines on the plot

        plot_kwargs : dict
            Key word arguments for the main plot, see `pyleoclim.Series.plot <https://pyleoclim-util.readthedocs.io/en/latest/core/api.html#series-pyleoclim-series>_ for details

        ci_kwargs : dict
            Key word arguments for calculating the confidence interval. Only to be used if `transition_interval` is not passed. See ammonyte.utils.sampling.confidence_interval for details

        background_series : pyleoclim.Series
            Optional to pass a different series that will be plotted behind the main series plot

        background_kwargs : dict
            Key word arguments for the background plot. If none are passed, the color of the main plot will be re-used 
            and alpha will be set to .2

        legend : bool; {True,False}
            Whether or not to plot the legend

        lgd_kwargs : dict

        Returns
        -------

        fig : matplotlib.figure
            The figure object from matplotlib.
            See `matplotlib.pyplot.figure <https://matplotlib.org/stable/api/figure_api.html>`_ for details.

        ax : matplotlib.axis 
            The axis object from matplotlib. 
            See `matplotlib.axes <https://matplotlib.org/stable/api/axes_api.html>`_ for details.


        See also
        --------

        ammonyte.utils.sampling.confidence_interval

        '''

        ci_kwargs = {} if ci_kwargs is None else ci_kwargs.copy()
        plot_kwargs = {} if plot_kwargs is None else plot_kwargs.copy()
        lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()

        series = self.interp(step=1)

        if not transition_interval:
            transition_interval = confidence_interval(series,**ci_kwargs)
        
        if line_color is None:
            line_color=sns.color_palette('colorblind')[0]
        if fill_color is None:
            fill_color=sns.color_palette('colorblind')[0]
        
        #Need to find points in time where the series intersects with the lower and upper confidence boundaries
        value = series.value
        time = series.time
        
        idx_ufill = []
        idx_lfill = []

        upper = max(transition_interval)
        lower = min(transition_interval)

        #Need to find all points that lie above or below our confidence interval
        for idx,v in enumerate(value):
            if v >= upper:
                idx_ufill.append(idx)
            elif v <= lower:
                idx_lfill.append(idx)
        
        if ax is None:
            fig,ax = plt.subplots(figsize=(12,8))

        if background_series:

            if not isinstance(background_series,pyleo.core.Series):
                raise ValueError('background_series does not appear to be a pyleoclim.Series object')
            
            #Standardize background_series values
            background_series.value -= np.mean(background_series.value)
            background_series.value /= max(background_series.value)

            #Adjust scaling to match that of main series
            background_series.value *= max(series.value - np.mean(series.value))
            background_series.value += np.mean(series.value)

            background_kwargs = {} if background_kwargs is None else background_kwargs.copy()

            if 'alpha' not in background_kwargs:
                background_kwargs['alpha'] = .2
            if 'color' not in background_kwargs:
                background_kwargs['color'] = line_color

            background_series.plot(ax=ax, **background_kwargs)

        if fill_alpha is None:
            fill_alpha = .1

        if xlabel is None:
            xlabel = f'{self.time_name} [{self.time_unit}]'

        if ylabel is None:
            ylabel = f'{self.value_name} [{self.value_unit}]'

        if label is None:
            label = self.label

        if label is not None:
            plot_kwargs.update({'label': label})

        if marker is not None:
            plot_kwargs.update({'marker': marker})

        if markersize is not None:
            plot_kwargs.update({'markersize': markersize})

        if linestyle is not None:
            plot_kwargs.update({'linestyle': linestyle})

        if linewidth is not None:
            plot_kwargs.update({'linewidth': linewidth})

        if alpha is not None:
            plot_kwargs.update({'alpha': alpha})

        if zorder is not None:
            plot_kwargs.update({'zorder': zorder})

        series.plot(ax=ax,color=line_color,xlabel=xlabel,ylabel=ylabel,title=title,plot_kwargs=plot_kwargs,lgd_kwargs=lgd_kwargs,legend=legend)
            
        ufill_values = np.zeros(len(time)) + upper
        lfill_values = np.zeros(len(time)) + lower
        
        for i in range(len(time)):
            if i in idx_ufill:
                ufill_values[i] = value[i]
            elif i in idx_lfill:
                lfill_values[i] = value[i]
                                    
        ax.fill_between(time,lower,upper,color=fill_color,alpha=fill_alpha)
        ax.fill_between(time,upper,ufill_values,color=fill_color)
        ax.fill_between(time,lower,lfill_values,color=fill_color)

        if 'fig' in locals():
            return fig, ax
        else:
            return ax

    def confidence_scatter_plot(self,ax=None,transition_interval=None,marker=None,color=None,size=None,
                    label=None,xlabel=None,ylabel=None,title=None,scatter_kwargs=None,ci_kwargs=None,
                    background_series=None,background_kwargs=None,legend=True,lgd_kwargs=None):

        '''Function for plotting rqa results with confidence bounds

        Parameters
        ----------

        ax : matplotlib.axes object
            Axes to plot on, if None new plot will be generated

        transition_interval : list,tuple
            Upper and lower bound for the transition interval

        marker : str
            e.g., 'o' for dots
            See `matplotlib.markers <https://matplotlib.org/stable/api/markers_api.html>`_ for details

        size : float
            the size of the marker

        color : str, tuple
            String or rgb tuple to use for marker color

        label : str
            the label for the scatter

        xlabel : str
            the label for the x-axis

        ylabel : str
            the label for the y-axis

        title : str
            the title for the figure

        scatter_kwargs : dict
            Key word arguments for the main plot, see `matplotlib.pyplot.scatter <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.scatter.html>_ for details

        ci_kwargs : dict
            Key word arguments for calculating the confidence interval. Only to be used if `transition_interval` is not passed. See ammonyte.utils.sampling.confidence_interval for details

        background_series : pyleoclim.Series
            Optional to pass a different series that will be plotted behind the main series plot

        background_kwargs : dict
            Key word arguments for the background plot. If none are passed, the color of the main plot will be re-used 
            and alpha will be set to .2

        legend : bool; {True,False}
            Whether or not to plot the legend

        lgd_kwargs : dict

        Returns
        -------

        fig : matplotlib.figure
            The figure object from matplotlib.
            See `matplotlib.pyplot.figure <https://matplotlib.org/stable/api/figure_api.html>`_ for details.

        ax : matplotlib.axis 
            The axis object from matplotlib. 
            See `matplotlib.axes <https://matplotlib.org/stable/api/axes_api.html>`_ for details.


        See also
        --------

        ammonyte.utils.sampling.confidence_interval

        '''

        ci_kwargs = {} if ci_kwargs is None else ci_kwargs.copy()
        scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs.copy()
        lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()

        if not transition_interval:
            transition_interval = confidence_interval(self,**ci_kwargs)
        
        idx_fill = []

        upper = max(transition_interval)
        lower = min(transition_interval)

        #Need to find all points that lie above or below our confidence interval
        for idx,v in enumerate(self.value):
            if v >= upper or v <= lower:
                idx_fill.append(idx)
        
        if ax is None:
            fig,ax = plt.subplots(figsize=(12,8))

        if background_series:

            if not isinstance(background_series,pyleo.core.Series):
                raise ValueError('background_series does not appear to be a pyleoclim.Series object')
            
            #Standardize background_series values
            background_series.value -= np.mean(background_series.value)
            background_series.value /= max(background_series.value)

            #Adjust scaling to match that of main series
            background_series.value *= max(self.value - np.mean(self.value))
            background_series.value += np.mean(self.value)

            background_kwargs = {} if background_kwargs is None else background_kwargs.copy()

            if 'alpha' not in background_kwargs:
                background_kwargs['alpha'] = .2

            background_series.plot(ax=ax, **background_kwargs)

        if xlabel is None:
            xlabel = f'{self.time_name} [{self.time_unit}]'

        if ylabel is None:
            ylabel = f'{self.value_name} [{self.value_unit}]'

        if label is None:
            label = self.label

        if marker is None:
            marker = 'x'
        
        if color is None:
            color = sns.color_palette('colorblind')[2]

        if size is None:
            size = 30

        ax.scatter(self.time[idx_fill],self.value[idx_fill],label=label,marker=marker,s=size,c=color,
                    **scatter_kwargs)

        ax.axhline(max(transition_interval),0,1,color=color)
        ax.axhline(min(transition_interval),0,1,color=color)

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title)

        if legend:
            ax.legend(**lgd_kwargs)

        if 'fig' in locals():
            return fig, ax
        else:
            return ax


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import itertools

import pyleoclim as pyleo
import numpy as np
import pandas as pd
import multiprocessing as mp

from pyrqa.time_series import EmbeddedSeries
from pyrqa.settings import Settings
from pyrqa.analysis_type import Classic
from pyrqa.neighbourhood import FixedRadius
from pyrqa.metric import EuclideanMetric
from pyrqa.computation import RPComputation

from ..core.recurrence_matrix import RecurrenceMatrix
from ..core.recurrence_network import RecurrenceNetwork
from ..utils.parameters import tau_search
from ..utils.range_finder import range_finder


class TimeEmbeddedSeries:
    '''Time embedded time series object. Precursor to recurrence matrix and recurrence network.

    series : pyleo.Series object or pandas.Series object
        Time series to be embedded
    
    m : int
        Embedding dimension
    
    tau : int
        Embedding delay, will be calculated according to first minimum of mutual information if not passed

    embedded_data : array
        Time delay embedded data. If not passed will be calculated from series, m, and tau

    embedded_time : array
        Time axis corresponding to embedded_data. If embedded_data is passed without embedded_time, an error will be raised

    value_name : str
        Value name

    value_unit : str
        Value units

    time_name : str
        Time name

    time_unit : str
        Time units

    label : str
        Label for embedding
    '''

    def __init__(self,series,m,tau=None,embedded_data=None,embedded_time=None,value_name=None,value_unit=None,time_name=None,time_unit=None,label=None):
        self.series = series
        self.m = m
        self.tau = tau
        self.embedded_data = embedded_data
        self.embedded_time = embedded_time
        self.value_name = value_name
        self.value_unit = value_unit
        self.time_name = time_name
        self.time_unit = time_unit
        self.label = label

        if self.embedded_data is not None and self.embedded_time is None:
            raise ValueError('Embedded data was passed without associated time axis. Please pass neither or both')

        if self.tau is None:
            self.tau = tau_search(self.series)

        if self.embedded_data is None:

            if isinstance(self.series, (pyleo.core.Series, pyleo.core.LipdSeries)):
                values = self.series.value
                time_axis = self.series.time[:(-self.m*self.tau)]

            elif isinstance(self.series,pd.Series):
                values = self.series.values
                time_axis = list(self.series.index)[:(-self.m*self.tau)]

            else:
                raise ValueError('Unrecognized data type. Please pass a pyleoclim Series or pandas Series type object')
            
            manifold = np.ndarray(shape = (len(values)-(self.m*self.tau),self.m))

            for idx, i in enumerate(values):
                if idx < (len(values)-(self.m*self.tau)):
                    manifold[idx] = values[idx:idx+(self.m*self.tau):self.tau]

            self.embedded_data = manifold
            self.embedded_time = time_axis

        if self.value_name is None:
            self.value_name = self.series.value_name
        
        if self.value_unit is None:
            self.value_unit = self.series.value_unit

        if self.time_name is None:
            self.time_name = self.series.time_name
        
        if self.time_unit is None:
            self.time_unit = self.series.time_unit
        
        if self.label is None:
            self.label = self.series.label

    def create_recurrence_matrix(self,epsilon):
        '''Function to create Recurrence Matrix object
        
        Parameters
        ----------
        
        epsilon : float
            Fixed radius used to calculate whether two points are recurrent
            
        Returns
        -------
        
        RecurrenceMatrix : ammonyte.RecurrenceMatrix object'''

        ts = EmbeddedSeries(self.embedded_data)

        settings = Settings(ts,
                            analysis_type=Classic,
                            neighbourhood=FixedRadius(epsilon),
                            similarity_measure=EuclideanMetric)

        computation = RPComputation.create(settings,
                                        verbose=False)

        result = computation.run()

        matrix = result.recurrence_matrix

        return RecurrenceMatrix(matrix=matrix,time=self.embedded_time,epsilon=epsilon,series=self.series, m = self.m, tau = self.tau,
                                value_name=self.value_name,value_unit=self.value_unit,time_name=self.time_name,
                                time_unit=self.time_unit,label=self.label)

    def create_recurrence_network(self,epsilon):
        '''Function to create Recurrence Network object
        
        Parameters
        ----------
        
        epsilon : float
            Fixed radius used to calculate whether two points are recurrent.
            
        Returns
        -------
        
        RecurrenceNetwork : ammonyte.RecurrenceNetwork object'''

        ts = EmbeddedSeries(self.embedded_data)

        settings = Settings(ts,
                            analysis_type=Classic,
                            neighbourhood=FixedRadius(epsilon),
                            similarity_measure=EuclideanMetric)

        computation = RPComputation.create(settings,
                                        verbose=False)

        result = computation.run()

        matrix = result.recurrence_matrix

        return RecurrenceNetwork(matrix=matrix,time=self.embedded_time,epsilon=epsilon,series=self.series,
                                value_name=self.value_name,value_unit=self.value_unit,time_name=self.time_name,
                                time_unit=self.time_unit,label=self.label)

    def find_epsilon(self,eps,target_density=.05,tolerance=.01,initial_density=None,parallelize=True,num_processes=None,amp=10,verbose=True):
        '''Function to find epsilon value given target recurrence matrix density
        
        Parameters
        ----------
        eps : float
            Starting epsilon value (best guess)
        target_density : float
            Desired recurrence matrix density
        tolerance : float
            Amount of allowable difference between target density and actual density
        initial_density : float
            If you've already calculated the initial density for your settings you can pass it here to save computation time
        parallelize : bool; {True,False}
            Whether or not to parallelize the search process. Currently only tested on macOS, could be issues running this on Windows.
        num_processes : int
            Number of processes to run, automatically set to your cpu count
        amp : int
            The amplitude of the range of epsilon value search. Higher values cover ground quickly but converge slowly, the opposite is true for lower values
        verbose : bool; {True,False}
            Whether or not to print output after each iteration
        Returns
        -------
        epsilon : float
            Epsilon value that produces desired recurrence density within tolerance value provided.
        See also
        --------
        ammonyte.utils.rm_search
        '''

        if num_processes is None:
            if mp.cpu_count() > 2:
                num_processes = mp.cpu_count() - 2
            else:
                num_processes = 1
        
        if initial_density is None:

            initial_result = self.create_recurrence_matrix(eps)
            initial_density = np.sum(initial_result.matrix)/np.size(initial_result.matrix)

            if verbose:
                print(f'Initial density is {initial_density:.4f}')
        
        if np.abs(initial_density - target_density) <= tolerance:

            if verbose:
                print('Initial density is within the tolerance window!')

            results = {'Epsilon':eps,'Output':initial_result}

            return results
        else:
            if verbose:
                print('Initial density is not within the tolerance window, searching...')
            density = initial_density

        if parallelize:

            while True:

                with mp.Pool(num_processes) as pool:
                    
                    eps_range, flag = range_finder(eps,density,target_density,tolerance,num_processes,amp)
                    
                    if flag is True:
                        
                        eps = eps_range
                        results = {'Epsilon':eps,'Output':self.create_recurrence_matrix(eps)}

                        if verbose:
                            matrix = results['Output'].matrix
                            density = np.sum(matrix)/np.size(matrix)
                            print(f'Epsilon: {eps:.4f}, Density: {density:.4f}.')

                        return results

                    r = pool.starmap(self.create_recurrence_matrix, zip(eps_range))
                    
                    pool.close()
                    pool.join()

                for item in r:
                    matrix = item.matrix
                    new_eps = item.epsilon
                    new_density = np.sum(matrix)/np.size(matrix)

                    if np.abs(new_density - .05) < np.abs(density -.05):
                        density = new_density
                        eps = new_eps

                if verbose:
        
                    print(f'Epsilon: {eps:.4f}, Density: {density:.4f}.')
        
        else:
            modifier=1
            while True:

                distance = target_density-density

                if distance <= tolerance:
                        
                        results = {'Epsilon':eps,'Output':self.create_recurrence_matrix(eps)}

                        if verbose:
                            matrix = results['Output'].matrix
                            density = np.sum(matrix)/np.size(matrix)
                            print(f'Epsilon: {eps:.4f}, Density: {density:.4f}.')

                        return results

                new_eps = eps+(amp*distance*modifier)
                trial = self.create_recurrence_matrix(new_eps)
                matrix = trial.matrix
                new_eps = trial.epsilon
                new_density = np.sum(matrix)/np.size(matrix)
                new_distance = target_density - new_density

                if np.abs(new_distance) <= np.abs(distance):
                    density = new_density
                    eps = new_eps
                    modifier=1

                elif np.abs(new_distance) > np.abs(distance):
                    modifier /= 2

                if verbose:
        
                    print(f'Epsilon: {eps:.4f}, Density: {density:.4f}.')



# -*- coding: utf-8 -*-
"""
N. Ahmad, S. Derrible, T. Eason, and H. Cabezas, 2016, “Using Fisher information to track stability in multivariate systems”,
Royal Society Open Science, 3:160582, DOI: 10.1098/rsos.160582

Edits have been made by Alexander James to streamline the code for usage in this package
"""

import pandas as pd 
import math
import numpy as np
import pyleoclim as pyleo

__all__ = [
    'fisher_information',
    'smooth_series'
]

def fisher_information(eig_data,w_size,w_incre):
    Data_num=[]
    Time=[]
    
    for row in eig_data:
        Time.append(row[0])
        temp=[]
        for i in range(1,len(row)):
            if row[i]=='':
                temp.append(0)
            else:
                temp.append(float(row[i]))
        Data_num.append(temp)
        
    sost_data = SOST(eig_data,w_size)
    sost = sost_data.values[0]
    
    FI_final=[]
    k_init=[]
    for i in range(0,len(Data_num),w_incre):
        
        Data_win=Data_num[i:i+w_size]
        win_number=i
        
        if len(Data_win)==w_size:
            Bin=[]
            for m in range(len( Data_win)):
                Bin_temp=[]
                
                for n in range(len( Data_win)):
                    if m==n:
                        Bin_temp.append('I')
                    else:
                        Bin_temp_1=[]
                    
                        for k in range(len(Data_win[n])):
                            if (abs(Data_win[m][k]-Data_win[n][k]))<=sost[k]:
                                Bin_temp_1.append(1)
                            else:
                                Bin_temp_1.append(0)
                                
                        Bin_temp.append(sum(Bin_temp_1))
                        
                Bin.append(Bin_temp)
            
            FI=[]
            for tl in range(1,101):
                tl1=len(sost)*float(tl)/100
                Bin_1=[]
                Bin_2=[]
                
                for j in range(len(Bin)):
                    if j not in Bin_2:
                       
                        Bin_1_temp=[j]
                        for i in range(len(Bin[j])):
                            if Bin[j][i]!='I' and Bin[j][i]>=tl1 and i not in Bin_2:
                                Bin_1_temp.append(i)
                                
                        Bin_1.append(Bin_1_temp)
                        Bin_2.extend(Bin_1_temp)

                prob=[0]
                for i in Bin_1:
                    prob.append(float(len(i))/len(Bin_2))
                    
                prob.append(0)
                
                prob_q=[]
                for i in prob:
                    prob_q.append(math.sqrt(i))
                    
                FI_temp=0
                for i in range(len(prob_q)-1):
                    FI_temp+=(prob_q[i]-prob_q[i+1])**2
                FI_temp=4*FI_temp    
                
                FI.append(FI_temp)
                
            for i in range(len(FI)):
                if FI[i]!=8.0:
                    k_init.append(FI.index(FI[i]))
                    break
                
            FI_final.append(FI)
            
    if len(k_init)==0:
        k_init.append(0)
        
    time_axis = []
    
    for i in range(0,len(FI_final)):
        FI_final[i].append(float(sum(FI_final[i][min(k_init):len(FI_final[i])]))/len(FI_final[i][min(k_init):len(FI_final[i])]))
        time_axis.append(Time[(i*w_incre+w_size)-1])
        
    FI_final = pd.DataFrame(FI_final)
    values = FI_final.iloc[:,-1]
    
    return np.array(time_axis), np.array(values)
        
def SOST(eig_data,s_for_sd):
    Data_num=[]
    
    for row in eig_data:
        temp=[]
        for i in range(1,len(row)):
            if row[i]=='':
                temp.append(0)
            else:
                temp.append(float(row[i]))
        Data_num.append(temp)
        
    df=pd.DataFrame(Data_num)
    
    sos=[]
    for j in range(len(df.columns)):
        sos_temp=[]
        for i in df.index:
            A=list(df[j][i:i+s_for_sd])
            A_1=[float(i) for i in A if i!=0 ]
        
            if len(A_1)==s_for_sd:
                sos_temp.append(np.std(A_1,ddof=1))
                
        if len(sos_temp)==0:
            sos.append(0)
        else:
            
            sos.append(min(sos_temp)*2)
        
    df_sos=pd.DataFrame(sos)
    df_sos=df_sos.transpose()
    
    return df_sos
    
def smooth_series(series,block_size):
    '''Function to smooth series by averaging chunks
    
    Parameters
    ----------
    
    series : pyleoclim.Series object
        Series to be smoothed
        
    block_size : int
        Size of each block, assumes an evenly spaced series'''
    
    #There is definitely a better way to do this, will fix in the future
    
    values = series.value
    
    smoothed_values=[]
   
    for i in range(block_size,len(values)+block_size,block_size):
        for _ in range(i-block_size,i):
            smoothed_values.append(float(sum(values[i-block_size:i]))/len(values[i-block_size:i]))
            
    smoothed_values=smoothed_values[0:len(values)]
    
    series.value = np.array(smoothed_values)
    
    return series


import pyleoclim as pyleo
import numpy as np
import pandas as pd

from pyrqa.time_series import TimeSeries
from pyrqa.settings import Settings
from pyrqa.analysis_type import Classic
from pyrqa.neighbourhood import FixedRadius
from pyrqa.metric import EuclideanMetric
from pyrqa.computation import RPComputation
    
def rm(series, eps, m, delay):
    '''Function to calculate recurrence matrix from pyleoclim or pandas series
    
    Parameters
    ----------
    
    series : pyleoclim.Series or pandas.Series
        Timeseries from which the recurrence matrix will be calculated
    eps : int
        Epsilon value for calculation of the matrix
    m : int
        Embedding parameter
    delay : int
        Delay parameter for time embedding
    '''
    
    values = series.value
    time_axis = series.time[:-(m-1)*delay]
        
    ts = TimeSeries(values,
                    embedding_dimension = m,
                    time_delay=delay)

    settings = Settings(ts,
                        analysis_type=Classic,
                        neighbourhood=FixedRadius(eps),
                        similarity_measure=EuclideanMetric)

    computation = RPComputation.create(settings,
                                        verbose=False)

    result = computation.run()

    matrix = result.recurrence_matrix

    return {'rm':matrix, 'time_axis':time_axis,'eps':eps}


''' Tests for ammonyte.core.series
Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
    (certain tests will only work when run from the tests directory, so make sure to run from there!)
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''

import pytest
import pyleoclim as pyleo
import ammonyte as amt
import numpy as np

def gen_normal(loc=0, scale=1, nt=100):
    ''' Generate random data with a Gaussian distribution
    '''
    t = np.arange(nt)
    np.random.seed(42)
    v = np.random.normal(loc=loc, scale=scale, size=nt)
    ts = amt.Series(t,v)
    return ts

class TestCoreSeriesEmbed:
    '''Tests for embed function
    '''

    @pytest.mark.parametrize('m,tau',[(10,5),(10,None)])
    def test_embed_t0(self,m,tau):
        '''Test embed function with and without a tau value'''

        ts = gen_normal()

        ts.embed(m,tau)

class TestCoreSeriesDeterminism:
    '''Tests for determinism function
    '''

    @pytest.mark.parametrize('window_size,overlap,radius,m,tau',[(10,5,1,5,2),(12,4,.1,8,4)])
    def test_determinism_t0(self,window_size,overlap,m,tau,radius):

        ts = gen_normal()

        ts.determinism(window_size,overlap,m,tau,radius)

class TestCoreSeriesLaminarity:
    '''Tests for laminarity function'''

    @pytest.mark.parametrize('window_size,overlap,radius,m,tau',[(10,5,1,5,2),(12,4,.1,8,4)])
    def test_laminarity_t0(self,window_size,overlap,m,tau,radius):

        ts = gen_normal()

        ts.laminarity(window_size,overlap,m,tau,radius)


''' Tests for ammonyte.core.rqa_res
Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
    (certain tests will only work when run from the tests directory, so make sure to run from there!)
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''

import pytest
import ammonyte as amt
import numpy as np

def gen_normal(loc=0, scale=1, nt=100):
    ''' Generate random data with a Gaussian distribution
    '''
    t = np.arange(nt)
    np.random.seed(42)
    v = np.random.normal(loc=loc, scale=scale, size=nt)
    ts = amt.Series(t,v)
    return ts

class TestCoreRQAResSmooth:
    '''Tests for smooth function'''

    @pytest.mark.parametrize('block_size',(5,None))
    def test_smooth_t0(self,block_size):
        ts_normal = gen_normal()
        amt_td = amt.TimeEmbeddedSeries(ts_normal,3,1)
        rm = amt_td.create_recurrence_matrix(1)
        lp_series = rm.laplacian_eigenmaps(5,3)
        lp_series.smooth(block_size)

class TestCoreRQAResConfidenceScatterPlot:
    '''Tests for confidence scatter plot function'''

    @pytest.mark.parametrize('marker,size,color,legend,label,xlabel,ylabel,title,scatter_kwargs,lgd_kwargs',
                            [('o',42,'red',True,'label_test','xlabel_test','ylabel_test','title_test',{'alpha':.5},{'fontsize':42}),
                            (None,None,None,False,None,None,None,None,None,None)])
    def test_confidence_scatter_plot_t0(self,marker,size,color,legend,label,xlabel,ylabel,title,scatter_kwargs,lgd_kwargs):
        '''Testing different visual plot arguments'''
        #Parameter choices are completely arbitrary, just want to test if the plotting function works
        ts_normal = gen_normal(nt=100)
        amt_td = amt.TimeEmbeddedSeries(ts_normal,3,1)
        rm = amt_td.create_recurrence_matrix(1)
        lp_series = rm.laplacian_eigenmaps(5,3)
        lp_series.confidence_scatter_plot(marker=marker,size=size,color=color,legend=legend,label=label,xlabel=xlabel,ylabel=ylabel,title=title,
                                          scatter_kwargs=scatter_kwargs,lgd_kwargs=lgd_kwargs)

    @pytest.mark.parametrize('transition_interval,ci_kwargs',([(0,1),None],[(1,-1),None],[None,{'upper':75,'lower':15}]))
    def test_confidence_scatter_plot_t1(self,transition_interval,ci_kwargs):
        '''Testing different confidence interval calculations'''
        #Parameter choices are completely arbitrary, just want to test if the plotting function works
        ts_normal = gen_normal(nt=100)
        amt_td = amt.TimeEmbeddedSeries(ts_normal,3,1)
        rm = amt_td.create_recurrence_matrix(1)
        lp_series = rm.laplacian_eigenmaps(5,3)
        lp_series.confidence_scatter_plot(transition_interval=transition_interval,ci_kwargs=ci_kwargs)

    @pytest.mark.parametrize('background_kwargs',(None,{'alpha':1}))
    def test_confidence_scatter_plot_t2(self,background_kwargs):
        '''Testing with background plot'''
        #Parameter choices are completely arbitrary, just want to test if the plotting function works
        ts_normal = gen_normal(nt=100)
        amt_td = amt.TimeEmbeddedSeries(ts_normal,3,1)
        rm = amt_td.create_recurrence_matrix(1)
        lp_series = rm.laplacian_eigenmaps(5,3)
        lp_series.confidence_scatter_plot(background_series=ts_normal,background_kwargs=background_kwargs)

class TestCoreRQAResConfidenceFillPlot:
    '''Tests for confidence fill plot function'''

    @pytest.mark.parametrize('line_color,fill_color,fill_alpha,legend,label,xlabel,ylabel,title,plot_kwargs,lgd_kwargs',
                            (['green','purple',1,True,'label_test','xlabel_test','ylabel_test','title_test',{'alpha':.5},{'fontsize':42}],
                            [None,None,None,False,None,None,None,None,None,None]))
    def test_confidence_fill_plot_t0(self,line_color,fill_color,fill_alpha,legend,label,xlabel,ylabel,title,plot_kwargs,lgd_kwargs):
        '''Testing different visual plot arguments'''
        #Parameter choices are completely arbitrary, just want to test if the plotting function works
        ts_normal = gen_normal(nt=100)
        amt_td = amt.TimeEmbeddedSeries(ts_normal,3,1)
        rm = amt_td.create_recurrence_matrix(1)
        lp_series = rm.laplacian_eigenmaps(5,3)
        lp_series.confidence_fill_plot(line_color=line_color,fill_color=fill_color,fill_alpha=fill_alpha,
                                       legend=legend,label=label,xlabel=xlabel,ylabel=ylabel,title=title,plot_kwargs=plot_kwargs,
                                       lgd_kwargs=lgd_kwargs )
        
    @pytest.mark.parametrize('transition_interval,ci_kwargs',([(0,1),None],[(1,-1),None],[None,{'upper':75,'lower':15}]))
    def test_confidence_fill_plot_t1(self,transition_interval,ci_kwargs):
        '''Testing different confidence interval calculations'''
        #Parameter choices are completely arbitrary, just want to test if the plotting function works
        ts_normal = gen_normal(nt=100)
        amt_td = amt.TimeEmbeddedSeries(ts_normal,3,1)
        rm = amt_td.create_recurrence_matrix(1)
        lp_series = rm.laplacian_eigenmaps(5,3)
        lp_series.confidence_fill_plot(transition_interval=transition_interval,ci_kwargs=ci_kwargs)

    @pytest.mark.parametrize('background_kwargs',(None,{'alpha':1}))
    def test_confidence_fill_plot_t2(self,background_kwargs):
        '''Testing with background plot'''
        #Parameter choices are completely arbitrary, just want to test if the plotting function works
        ts_normal = gen_normal(nt=100)
        amt_td = amt.TimeEmbeddedSeries(ts_normal,3,1)
        rm = amt_td.create_recurrence_matrix(1)
        lp_series = rm.laplacian_eigenmaps(5,3)
        lp_series.confidence_fill_plot(background_series=ts_normal,background_kwargs=background_kwargs)


import multiprocessing as mp
import itertools
import pyleoclim as pyleo
import numpy as np

import pyinform

from scipy.signal import argrelextrema
from tqdm import tqdm

from ..utils.rm import rm
from ..utils.range_finder import range_finder
# from ..core.time_embedded_series import TimeEmbeddedSeries


__all__ = [
    'tau_search',
    'eps_search'
]

def tau_search(series,num_lags=30,return_MI = False):
    '''Find optimal tau value for time delay embedding.
    
    First minimum of mutual information between series and time lagged copies of itself
    is "optimal" in this case in accordance with Abarnabel's "Analysis of Observed Chaotic Data"

    Parameters
    ----------

    series : pyleo.Series
        Series for which we'd like to find the optimal tau value

    num_lags : int
        Number of time delays to consider. Default is 30

    return_MI : bool, {True,False}
        Whether or not to return the list of mutual information values. 
        Useful if the first minimum seems spurious and you'd like to inspect the results.

    Returns
    -------

    tau : int
        Optimal time delay parameter according to first minimum of mutual information

    MI : list
        List of mutual information values.
        Indices + 1 correspond to amount of lag (index 0 is lag 1, index 1 is lag 2, etc.).
        Only returned if return_MI is set to True.
    
    Citations
    ---------
    
    I., Abarbanel Henry D. Analysis of Observed Chaotic Data. Springer, 1997. 
    '''
    lags = np.arange(1,num_lags)
    MI = []

    for lag in lags:
        values = series.value[:-lag] - min(series.value[:-lag])
        lagged_values = series.value[lag:] - min(series.value[lag:])
        MI.append(pyinform.mutualinfo.mutual_info(values, lagged_values, local=False))

    best_tau = argrelextrema(np.array(MI),np.less)[0][0] + 1

    if return_MI is True:
        return best_tau,MI
    else:
        return best_tau

def eps_search(series, m, tau ,target_density, tolerance, eps=1, amp = 15, initial_hitrate = None, num_processes = None,verbose = True):
    '''Tool to find epsilon value tuned for specific target density in recurrence matrix
    
    Parameters
    ----------
    
    series : pyleoclim.series object (pandas.series support incoming)
        Timeseries used to create recurrence matrix

    eps : float
        Starting epsilon value (best guess)

    m : int
        Embedding parameter for time delay embedding

    tau : int
        Delay parameter for time delay embedding

    target_density : float
        Desired recurrence matrix hitrate

    tolerance : float
        Amount of allowable difference between target hitrate and actual hitrate

    initial_hitrate : float
        If you've already calculated the initial hitrate for your settings you can pass it here to save computation time

    num_processes : int
        Number of processes to run, automatically set to your cpu count

    amp : int
        The amplitude of the range of epsilon value search. Higher values cover ground quickly but converge slowly, the opposite is true for lower values
        
    verbose : bool; {True,False}
        Whether or not to print output after each iteration
    '''
    
    if num_processes is None:
        if mp.cpu_count() > 2:
            num_processes = mp.cpu_count() - 2
        else:
            num_processes = 1
    
    if initial_hitrate == None:

        initial_result = rm(series, eps, m, tau)
        initial_hitrate = np.sum(initial_result['rm'])/np.size(initial_result['rm'])
        if verbose:
            print(f'Initial hitrate is {initial_hitrate:.4f}')
    
    if np.abs(initial_hitrate - target_density) <= tolerance:
        if verbose:
            print('Initial hitrate is within the tolerance window!')
        results = {'Epsilon':eps,'Output':initial_result}
        return results
    else:
        if verbose:
            print('Initial hitrate is not within the tolerance window, searching...')
        hitrate = initial_hitrate
        flag = True

    while flag:

        with mp.Pool(num_processes) as pool:
            
            eps_range, flag = range_finder(eps,hitrate,target_density,tolerance,num_processes,amp,verbose)
            
            if flag == False:
                
                eps = eps_range
                results = {'Epsilon':eps,'Output':rm(series, eps, m, tau)}
                return results
            
            r = pool.starmap(rm, zip(itertools.repeat(series), eps_range, itertools.repeat(m), itertools.repeat(tau)))
            
            pool.close()
            pool.join()

        if flag is True:
            for item in r:
                matrix = item['rm']
                new_eps = item['eps']
                new_hitrate = np.sum(matrix)/np.size(matrix)

                if np.abs(new_hitrate - .05) < np.abs(hitrate -.05):
                    hitrate = new_hitrate
                    eps = new_eps

        else:
            continue
    
    return results

# def grid_search(series, method, parameter_dict):
#     '''Function to apply a method with large number of parameters. Returns a collection of series objects with their associated parameters
    
#     Parameters
#     ----------
    
#     series : pyleoclim.Series or ammonyte.Series
#         Series to apply grid_search to
        
#     method : str
#         Method to apply. Current options include:
        
#         - laplacian_eigenmaps
#         - determinism
#         - laminarity

#     parameter_dict : dict
#         Dictionary of parameters to apply. Parameters should be included as keys with lists or arrays of parameter values as values

#     Returns
#     -------

#     res : list
#         List of ammonyte.RQA_Res objects
        
#     '''

#     if method == 'laplacian_eigenmaps':
#         keys, values = zip(*parameter_dict.items())
#         permutations_dicts = [dict(zip(keys, v)) for v in itertools.product(*values)]   

#         series_list = []

#         for permutation in tqdm(permutations_dicts):
#             series = series.bin(bin_size=permutation['bin_size']).detrend()
#             series_td = TimeEmbeddedSeries(series,permutation['m'],permutation['tau'])

#             if eps not in permutation:
#                 eps = series_td.find_epsilon(.05,.01,search_kwargs={'amp':50},verbose=False)

#             series_rm = series_td.create_recurrence_matrix(eps['Epsilon'])
#             lp_series = series_rm.laplacian_eigenmaps(50,5,smooth=False)
#             series_list.append(lp_series)

#         return series_list


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..core.recurrence_matrix import RecurrenceMatrix

class RecurrenceNetwork(RecurrenceMatrix):
    '''Recurrence network object. Used for Recurrence Network Analysis (RNA).

    Child of ammonyte.RecurrenceMatrix, so it has all of the RecurrenceMatrix methods
    plus additional methods defined here
    '''
    def __init__(self,matrix,time,epsilon,series=None,value_name=None,value_unit=None,time_name=None,time_unit=None,label=None):
        self.matrix = matrix
        self.time = time
        self.epsilon = epsilon
        self.series = series
        self.value_name = value_name
        self.value_unit = value_unit
        self.time_name = time_name
        self.time_unit = time_unit
        self.label = label


import numpy as np

def range_finder(eps, density, target_density, tolerance, num_processes, amp):
    
    if density < (target_density - tolerance):
        
        miss = target_density - density
        eps_bounds = (eps, eps + miss*amp)
        eps_range = np.linspace(min(eps_bounds), max(eps_bounds), num_processes)
        flag = False

        return eps_range, flag
    
    elif density > (target_density + tolerance):
        
        miss = density - target_density
        eps_bounds = (eps, eps - miss*amp)    
        eps_range = np.linspace(min(eps_bounds), max(eps_bounds), num_processes)      
        flag = False
        
        return eps_range, flag
        
    else:
            
        flag = True
        
        return eps, flag


import os
import sys
import io

from setuptools import setup, find_packages

version = '0.0.4'

# Read the readme file contents into variable
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='ammonyte',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    version=version,
    license='GPL-3.0 License',
    description='A Python package for nonlinear paleoclimate data analysis',
    long_description=read("README.md"),
    long_description_content_type = 'text/markdown',
    author='Alexander James',
    author_email='akjames@usc.edu',
    url='https://github.com/alexkjames/Ammonyte',
    keywords=['Paleoclimate, Data Analysis, Nonlinear'],
    classifiers=[],
    install_requires=[
        "pyleoclim>=0.7.0",
        "scipy>=1.7.1",
        "numpy>=1.21.5",
        "PyRQA>=8.0.0",
        "pyinform>=0.2.0"
    ],
    python_requires=">=3.8.0"
)


import numpy as np
from scipy.stats import scoreatpercentile
    
def confidence_interval(series,upper=95,lower=5,w=50,n_samples=10000,random_state = 42):
    '''Function to calculate upper and lower values for passed confidence interval on series object via bootstrapping
       Designed to be used to conduct bootstrap testing on fisher information series

    fisher_information : pyleoclim.Series
        Series to be evaluated

    upper : int
        Upper bound on confidence interval

    w : int
        Size of random sample
    
    n_samples : int
        Number of random samples
    
    random_state : int
        Random state for sampling
    '''
    
    rng = np.random.RandomState(seed=random_state)

    sub_arrays = []
    values = list(series.value)
    
    for i in range(n_samples):
        subset = rng.choice(values,size = w,)
        sub_arrays.append(subset)
    
    means = []

    for sequence in sub_arrays:
        means.append(np.mean(sequence))
    
    upper_val = scoreatpercentile(means, upper)
    lower_val = scoreatpercentile(means, lower)
    
    return upper_val, lower_val


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .recurrence_matrix import RecurrenceMatrix
from .recurrence_network import RecurrenceNetwork
from .time_embedded_series import TimeEmbeddedSeries
from .series import Series
from .rqa_res import RQARes


''' Tests for ammonyte.core.time_embedded_series
Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
    (certain tests will only work when run from the tests directory, so make sure to run from there!)
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''

import pytest
import pyleoclim as pyleo
import ammonyte as amt
import numpy as np

def load_data():
    #Loads stott MD982176 record
    d = pyleo.Lipd('../example_data/MD982176.Stott.2004.lpd')
    return d

def gen_normal(loc=0, scale=1, nt=100):
    ''' Generate random data with a Gaussian distribution
    '''
    t = np.arange(nt)
    np.random.seed(42)
    v = np.random.normal(loc=loc, scale=scale, size=nt)
    ts = amt.Series(t,v)
    return ts

class TestCoreTimeEmbeddSeriesCreateRecurrenceMatrix:
    '''Tests for create_recurrence_matrix
    '''

    @pytest.mark.parametrize('m,tau',[(3,3)])
    def test_create_recurrence_matrix_t0(self,m,tau):
        ts_normal = gen_normal()

        td_sst = ts_normal.embed(m,tau)

        td_sst.create_recurrence_matrix(1)

class TestCoreTimeEmbeddSeriesCreateRecurrenceNetwork:
    '''Tests for create_recurrence_network
    '''

    @pytest.mark.parametrize('m,tau',[(3,3)])
    def test_create_recurrence_network_t0(self,m,tau):
        ts_normal = gen_normal()

        td_sst = ts_normal.embed(m,tau)

        td_sst.create_recurrence_network(1)

class TestCoreTimeEmbeddSeriesFindEpsiilon:
    '''Tests for find_epsilon
    '''
    def test_find_eps_t0(self):
        ts_normal = gen_normal()

        td_sst = ts_normal.embed(3,1)

        td_sst.find_epsilon(1)